program example;

uses
  SysUtils, Classes,
  PEImage;

const
  LAYERSIZE = 44;               //  
  DATAOFS   = LAYERSIZE - 16;   //   (   ,   )

//  
procedure template;
asm
  { 60 } pushad
  { 9c } pushfd

  db $e8,0,0,0,0 // call +0
  { 5d } pop ebp

@next1:
  { 83 c5 xx } add ebp, DATAOFS  // ebp = data*

  mov ecx, [ebp+5]  // size
  mov esi, [ebp]    // va
  mov edi, esi

  cld

  // unpack
@loop1:

  lodsb

  xor al, $12
  rol al, $34
  add al, $56

  stosb
  dec ecx

  jnz @loop1

  // restore
  popfd
  popad

  push $FFFFFFFF // unpva
  ret
  dd $aaaaaaaa // unp size

end;

type
  // ,  
  TLayer = record
    p: pointer;             //    
    size: integer;          //  
    vxor, vrol, vadd: byte; //   
  end;

//  :
// UnpackVa:         ,   
// Size:             
// vxor,vrol,vadd:    
procedure createlayer(UnpackVa, size: uint32; vxor, vrol, vadd: byte; out layer: TLayer);
var
  p: pbyte;
begin
  layer.size := LAYERSIZE;
  getmem(layer.p, layer.size);
  p := layer.p;
  move(template, p^, LAYERSIZE);
  p[22] := vxor;
  p[25] := vrol;
  p[27] := vadd;
  pcardinal(@p[LAYERSIZE - 9])^ := UnpackVa;
  pcardinal(@p[LAYERSIZE - 4])^ := size;
end;

//     
procedure encode(p: pbyte; size: integer; vxor, vrol, vadd: byte);
var
  v: byte;
begin
  while size > 0 do
  begin
    v := p^;
    asm
      mov cl, vadd
      sub v, cl

      mov cl, vrol
      ror v, cl

      mov cl, vxor
      xor v, cl
    end;
    p^ := v;

    dec(size);
    inc(p);
  end;
end;

const
  max_layers = 100; //   = 100 ( + 1  )

var
  layers: array [0 .. max_layers] of TLayer;    //  ""
  pe: TPEImage;
  img: TMemoryStream;
  vamem: pointer;
  va, va0, va1, vaCaller: uint32;
  size: integer;
  vxor, vrol, vadd: byte;
  i: integer;
  sec: TPESectionHeader;

  p: pbyte;

begin
  va0 := $510774;         //     
  va1 := $510828;         //      
  vaCaller := $510754;    //   call  ,   

  pe := TPEImage.Create;        // PE 
  img := TMemoryStream.Create;  //  
  try
    if pe.LoadFromFile('puzzle.exe') then //  PE  
    begin
      //   1 ,       ()
      sec := pe.SecHdrs[0];
      sec.Flags := sec.Flags or IMAGE_SCN_MEM_WRITE;
      pe.SecHdrs[0] := sec;

      //   "megaprot"   max_layers + 1 
      // virtual size = raw size = size, flags = 0xE0000020
      size := (max_layers + 1) * LAYERSIZE;
      i := pe.AddSection('megaprot', 0, size, size, nil, $E0000020);
      sec := pe.SecHdrs[i];

      randomize;

      //   ,
      for i := 0 to max_layers do
      begin
        //   ""
        vxor := random(256);
        vrol := random(256);
        vadd := random(256);

        layers[i].vxor := vxor;
        layers[i].vrol := vrol;
        layers[i].vadd := vadd;

        if i = 0 then
        begin
          //    
          size := va1 - va0;
          createlayer(va0, size, vxor, vrol, vadd, layers[0]);
          vamem := pe.VAToMem(va0);
          encode(vamem, size, vxor, vrol, vadd); //  
        end
        else
        begin
          //    ,  
          va := sec.RVA + pe.ImageBase + (i - 1) * LAYERSIZE;
          createlayer(va, layers[i - 1].size, vxor, vrol, vadd, layers[i]);
        end;

      end;

      //       
      p := sec.Mem;
      for i := 0 to max_layers do
      begin
        move(layers[i].p^, p^, LAYERSIZE);
        if i <> max_layers then
          encode(p, layers[i].size, layers[i + 1].vxor, layers[i + 1].vrol, layers[i + 1].vadd);
        inc(p, LAYERSIZE);
      end;

      // va        ,
      va := sec.RVA + pe.ImageBase + max_layers * LAYERSIZE;

      //       (va)
      vamem := pe.VAToMem(vaCaller);
      inc(pbyte(vamem));                          //  call (E8)
      pinteger(vamem)^ := (va - (vaCaller + 5));  // 

      //        (  ,    )
      pe.BuildImage(img);

      //    
      img.SaveToFile('prot1.exe');

    end;
  finally
    img.Free;
    pe.Free;
  end;

end.
